# Copyright (c) 2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""I/O utilities for Pelican.

This module requires the :doc:`requests-pelican <requests-pelican:index>` package.

This module handles translation of ``pelican://`` URLs into lists of HTTP URLs
to pass back to `astropy.utils.data.download_file` and friends, and some
SciToken authorisation initialisation.
"""

from __future__ import annotations

import logging
from contextlib import contextmanager
from functools import (
    cache,
    wraps,
)
from typing import TYPE_CHECKING
from unittest import mock

from astropy.utils import data as astropy_data

from . import scitokens as io_scitokens

if TYPE_CHECKING:
    from collections.abc import Iterator
    from contextlib import AbstractContextManager
    from typing import (
        Any,
        BinaryIO,
        Literal,
    )

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

__all__ = [
    "download_file",
    "is_pelican_url",
    "open_remote_file",
    "query_director",
    "resolve_pelican_url",
]

log = logging.getLogger(__name__)


def _pelicanfs_protocols() -> list[str]:
    """Get list of known Pelican protocols from `pelicanfs`."""
    from pelicanfs.core import PelicanFileSystem
    return [klass.protocol for klass in [
        PelicanFileSystem,
        *PelicanFileSystem.__subclasses__(),
    ]]


def _requests_pelican_protocols() -> list[str]:
    """Get list of known Pelican protocols from `requests-pelican`."""
    from requests_pelican.federation import KNOWN_FEDERATIONS
    return ["pelican", *KNOWN_FEDERATIONS]


@cache
def _pelican_protocols() -> list[str]:
    """Return all known Pelican protocols.

    Needs `pelicanfs` or `requests-pelican`, otherwise returns an empty list.
    """
    for protocol_finder in (
        _pelicanfs_protocols,
        _requests_pelican_protocols,
    ):
        try:
            return protocol_finder()
        except ImportError:
            pass
    return []


@cache
def is_pelican_url(url: str) -> bool:
    """Return `True` if ``url`` looks like a Pelican URL.

    Needs the `pelicanfs` package, otherwise returns `False`.
    """
    try:
        protocols = list(_pelican_protocols())
    except ImportError:
        return False
    return any(url.startswith(f"{proto}://") for proto in protocols)


def query_director(
    url: str,
    **kwargs,
) -> tuple[list[str], bool, dict[str, Any]]:
    """Query a Pelican director for information.

    Parameters
    ----------
    url : `str`
        The Pelican URL of interest.

    kwargs
        All keyword arguments are passed to
        `requests_pelican.director.get_director_response`.

    Returns
    -------
    urls : `list` of `str`
        The list of HTTP URLs that serve the Pelican URL.

    need_auth : `bool`
        `True` if the HTTP URLs require token authorisation.

    auth_kw : `dict`
        Keyword arguments to pass to the
        `add_http_authorization_header` function to find/generate
        the appropriate token.

    See Also
    --------
    requests_pelican.director.get_director_response
    """
    from requests_pelican.director import get_director_response
    from urllib3.util import parse_url

    purl = parse_url(url)
    director = get_director_response(
        str(purl),
        **kwargs,
    )
    auth_kw: dict[str, Any] = {}
    if (
        # directory says we need auth
        (need_auth := director.namespace["require-token"])
        # but this is not an IGWN namespace (tokens are separately issued)
        and not director.namespace["namespace"].startswith("/igwn")
    ):
        for auth in director.auth:
            for key, val in auth.items():
                auth_kw.setdefault(key, []).append(val)
    return (
        director.urls(purl.path),
        need_auth,
        auth_kw,
    )


def resolve_pelican_url(
    url: str,
    **kwargs,
) -> list[str]:
    """Resolve a Pelican URL into one or more HTTP URLs."""
    return query_director(url, **kwargs)[0]


# -- astropy download hackery --------
#
# astropy.utils.data relies on an 'is_url' function to determine
# whether a target URL is an HTTP url. If that returns False,
# get_readable_filobj refuses to download a file, even if `sources`
# is given with a list of real HTTP URLs.
#
# So, we hack around it for Pelican.
#

@wraps(astropy_data.is_url)
def is_url(url: str) -> bool:
    """Return `True` for all Pelican URLs.

    If ``url`` doesn't look like a Pelican URL, hand off to
    `astropy.utils.data.is_url`.
    """
    if is_pelican_url(url):
        return True
    return astropy_data.is_url(url)


def _pelican_is_url_context() -> AbstractContextManager:
    return mock.patch("astropy.utils.data._is_url", is_url)


@contextmanager
def _pelican_download_context(
    reader: AbstractContextManager[BinaryIO],
) -> Iterator[BinaryIO]:
    """Wrap around ``reader`` to mock out `astropy.utils.data.is_url`."""
    with _pelican_is_url_context():  # noqa: SIM117
        with reader as result:
            yield result


def _prepare_download_kwargs(
    url: str,
    federation: str | None,
    cache: bool | Literal["update"],
    kwargs: dict[str, Any],
) -> tuple[str, dict[str, Any]]:
    """Prepare the keyword arguments for downloading a file from Pelican."""
    from requests_pelican.pelican import pelican_uri

    url = pelican_uri(url, federation=federation)

    # check if the file is already in the cache
    if cache is True:
        with _pelican_is_url_context():
            need_download = not astropy_data.is_url_in_cache(url)
    else:
        need_download = True

    # if we are going to download the file, query the Pelican director
    # for HTTP(S) sources, and required auth information
    if need_download:
        log.debug("Getting Pelican information from director")

        # query Pelican director for information
        sources, need_auth, auth_kw = query_director(url)

        # resolve URL to real HTTP URLs
        kwargs.setdefault("sources", sources)
        log.debug(
            "Resolved %d HTTP sources for %s",
            len(sources),
            url.rsplit("/", 1)[-1],
        )

        # add auth
        if need_auth:
            io_scitokens.add_http_authorization_header(
                kwargs.setdefault("http_headers", {}),
                url=url,
                error=False,
                create=False,
                **auth_kw,
            )

    return url, kwargs


def open_remote_file(
    url: str,
    *,
    federation: str | None = None,
    cache: bool | Literal["update"] = False,
    show_progress: bool = False,
    **kwargs,
) -> AbstractContextManager[BinaryIO]:
    """Download a remote file from a Pelican federation (and open it).

    This function is a wrapper around `astropy.utils.data.get_readable_fileobj`
    that uses :doc:`requests-pelican <requests-pelican:index>` to resolve the
    actual HTTP(S) URLs and pass those as the ``sources`` keyword to
    :func:`~astropy.utils.data.get_readable_fileobj`.

    Also, if the Pelican director informs that an authorisation token is
    required, this function attempts to an HTTP ``Authorization`` header
    using a locally-discovered `~scitokens.SciToken`
    (requires :doc:`igwn-auth-utils <igwn-auth-utils:index>`).

    Parameters
    ----------
    url : `str`
        The name of the resource to access. Must be a Pelican federation
        URL, either using the ``pelican://`` scheme or a federation
        scheme understood by `requests-pelican` (e.g. ``osdf://``), or
        the ``federation`` keyword must be provided.

    federation: `str`, optional
        The URL of the federation. This is required if ``url`` does not
        include federation information.

    cache : `bool`, ``"update"``, optional
        Whether to cache the contents of remote URLs.

    show_progress: `bool`, optional
        Print verbose progress information to the screen.

    kwargs
        All other positional and keyword arguments are passed directly
        to `astropy.utils.data.get_readable_fileobj`.

    Returns
    -------
    file : file-like
        The file opened in binary format.

    See Also
    --------
    astropy.utils.data.get_readable_fileobj
    """
    url, kwargs = _prepare_download_kwargs(
        url,
        federation,
        cache,
        kwargs,
    )

    # download the file using our 'is_url' hack
    # note: get_readable_fileobj is a context manager, so doesn't execute here
    return _pelican_download_context(astropy_data.get_readable_fileobj(
        url,
        cache=cache,
        show_progress=show_progress,
        **kwargs,
    ))


def download_file(
    url: str,
    *,
    federation: str | None = None,
    cache: bool | Literal["update"] = False,
    show_progress: bool = False,
    **kwargs,
) -> str:
    """Download a remote file from a Pelican federation.

    This function is a wrapper around `astropy.utils.data.download_file`
    that uses :doc:`requests-pelican <requests-pelican:index>` to resolve the
    actual HTTP(S) URLs and pass those as the ``sources`` keyword to
    :func:`~astropy.utils.data.download_file`.

    Also, if the Pelican director informs that an authorisation token is
    required, this function attempts to an HTTP ``Authorization`` header
    using a locally-discovered `~scitokens.SciToken`
    (requires :doc:`igwn-auth-utils <igwn-auth-utils:index>`).

    Parameters
    ----------
    url : `str`
        The name of the resource to access. Must be a Pelican federation
        URL, either using the ``pelican://`` scheme or a federation
        scheme understood by `requests-pelican` (e.g. ``osdf://``), or
        the ``federation`` keyword must be provided.

    federation: `str`, optional
        The URL of the federation. This is required if ``url`` does not
        include federation information.

    cache : `bool`, ``"update"``, optional
        Whether to cache the contents of remote URLs.

    show_progress: `bool`, optional
        Print verbose progress information to the screen.

    kwargs
        All other positional and keyword arguments are passed directly
        to `astropy.utils.data.download_file`.

    Returns
    -------
    file : file-like
        The file opened in binary format.

    See Also
    --------
    astropy.utils.data.download_file
    """
    url, kwargs = _prepare_download_kwargs(
        url,
        federation,
        cache,
        kwargs,
    )

    # download the file using our 'is_url' hack
    with _pelican_is_url_context():
        return astropy_data.download_file(
            url,
            cache=cache,
            show_progress=show_progress,
            **kwargs,
        )



# Copyright (c) 2017-2025 Cardiff University
#               2014-2017 Louisiana State University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""A python package for gravitational-wave astrophysics.

GWpy is a collaboration-driven `Python <http://www.python.org>`_ package
providing tools for studying data from ground-based gravitational-wave
detectors.

GWpy provides a user-friendly, intuitive interface to the common time-domain
and frequency-domain data produced by the `LIGO <http://www.ligo.org>`_ and
`Virgo <http://www.ego-gw.it>`_ instruments and their analysis,
with easy-to-follow tutorials at each step.
"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__credits__ = "The LIGO Scientific Collaboration and the Virgo Collaboration"

import logging

from . import (
    log,
    plot,  # registers gwpy.plot.Axes as default rectilinear axes
)
from .utils.env import bool_env

try:
    from ._version import version as __version__
except ModuleNotFoundError:  # development mode
    __version__ = ""


def init_logging(level: str | int = log.get_default_level()) -> None:
    """Quickly initialise logging for GWpy.

    This is mainly useful for scripts and interactive sessions to enable
    logging output from GWpy, and mainly for debugging purposes.

    For more control over logging, set up logging manually using the
    standard Python :mod:`logging` module, or see :ref:`gwpy-logging`.

    Parameters
    ----------
    level : int, optional
        The logging level to use.
        If not provided, the level from the ``GWPY_LOG_LEVEL`` environment
        variable will be used.
        If that variable is not set, ``INFO`` is used.

    Examples
    --------
    >>> import gwpy
    >>> gwpy.init_logging("DEBUG")
    """
    logger = log.init_logger(__name__, level=level or logging.INFO)
    logger.debug(
        "Initialised %s logging for %s",
        logging.getLevelName(logger.getEffectiveLevel()),
        logger.name,
    )


# Initialise logging for the package
if bool_env("GWPY_INIT_LOGGING", default=False):
    init_logging()
del bool_env



# Stubs for the SWIG-generated `lal` module used by GWpy.
# Minimal surface area: only symbols referenced by gwpy.utils.lal and related code.

# ruff: noqa: E501, N802, N803, N815

from collections.abc import Iterable
from typing import Generic, TypeVar, overload

import numpy

# Type code constants (integers)
I2_TYPE_CODE: int
I4_TYPE_CODE: int
I8_TYPE_CODE: int
U2_TYPE_CODE: int
U4_TYPE_CODE: int
U8_TYPE_CODE: int
S_TYPE_CODE: int
D_TYPE_CODE: int
C_TYPE_CODE: int
Z_TYPE_CODE: int

# -- Debug ---------------------------

def GetDebugLevel() -> int: ...
def ClobberDebugLevel(level: int) -> None: ...

# -- Cache ---------------------------

class Cache:
    length: int

def CacheGlob(dirname: str, pattern: str) -> Cache: ...

def CacheImport(path: str) -> Cache: ...

def CacheMerge(cache1: Cache, cache2: Cache) -> Cache: ...

# -- Detector-related types ----------

class FrDetector:
    name: str
    prefix: str
    vertexElevation: float
    vertexLatitudeRadians: float
    vertexLongitudeRadians: float
    xArmAltitudeRadians: float
    xArmAzimuthRadians: float
    xArmMidpoint: float
    yArmAltitudeRadians: float
    yArmAzimuthRadians: float
    yArmMidpoint: float

class DetectorContainer:
    frDetector: FrDetector

# CachedDetectors is an iterable of detector-like objects
CachedDetectors: Iterable[DetectorContainer]

# -- Vector --------------------------

class _Vector:
    length: int
    data: numpy.ndarray

_TVector = TypeVar("_TVector", bound=_Vector)

class INT2Vector(_Vector): ...
class INT4Vector(_Vector): ...
class INT8Vector(_Vector): ...
class UINT2Vector(_Vector): ...
class UINT4Vector(_Vector): ...
class UINT8Vector(_Vector): ...
class REAL4Vector(_Vector): ...
class REAL8Vector(_Vector): ...
class COMPLEX8Vector(_Vector): ...
class COMPLEX16Vector(_Vector): ...

def CreateINT2Vector(length: int) -> INT2Vector: ...
def CreateINT4Vector(length: int) -> INT4Vector: ...
def CreateINT8Vector(length: int) -> INT8Vector: ...
def CreateUINT2Vector(length: int) -> UINT2Vector: ...
def CreateUINT4Vector(length: int) -> UINT4Vector: ...
def CreateUINT8Vector(length: int) -> UINT8Vector: ...
def CreateREAL4Vector(length: int) -> REAL4Vector: ...
def CreateREAL8Vector(length: int) -> REAL8Vector: ...
def CreateCOMPLEX8Vector(length: int) -> COMPLEX8Vector: ...
def CreateCOMPLEX16Vector(length: int) -> COMPLEX16Vector: ...

# -- Series --------------------------

class _TimeSeries(Generic[_TVector]):
    data: _TVector
    deltaT: float
    epoch: LIGOTimeGPS
    f0: float
    length: int
    name: str
    sampleUnits: Unit

class INT2TimeSeries(_TimeSeries[INT2Vector]): ...
class INT4TimeSeries(_TimeSeries[INT4Vector]): ...
class INT8TimeSeries(_TimeSeries[INT8Vector]): ...
class UINT2TimeSeries(_TimeSeries[UINT2Vector]): ...
class UINT4TimeSeries(_TimeSeries[UINT4Vector]): ...
class UINT8TimeSeries(_TimeSeries[UINT8Vector]): ...
class REAL4TimeSeries(_TimeSeries[REAL4Vector]): ...
class REAL8TimeSeries(_TimeSeries[REAL8Vector]): ...
class COMPLEX8TimeSeries(_TimeSeries[COMPLEX8Vector]): ...
class COMPLEX16TimeSeries(_TimeSeries[COMPLEX16Vector]): ...

def CreateINT2TimeSeries(
    name: str | None, epoch: LIGOTimeGPS, f0: float,
    deltaT: float, sampleUnits: Unit, length: int,
) -> INT2TimeSeries: ...

def CreateINT4TimeSeries(
    name: str | None, epoch: LIGOTimeGPS, f0: float,
    deltaT: float, sampleUnits: Unit, length: int,
) -> INT4TimeSeries: ...

def CreateINT8TimeSeries(
    name: str | None, epoch: LIGOTimeGPS, f0: float,
    deltaT: float, sampleUnits: Unit, length: int,
) -> INT8TimeSeries: ...

def CreateUINT2TimeSeries(
    name: str | None, epoch: LIGOTimeGPS, f0: float,
    deltaT: float, sampleUnits: Unit, length: int,
) -> UINT2TimeSeries: ...

def CreateUINT4TimeSeries(
    name: str | None, epoch: LIGOTimeGPS, f0: float,
    deltaT: float, sampleUnits: Unit, length: int,
) -> UINT4TimeSeries: ...

def CreateUINT8TimeSeries(
    name: str | None, epoch: LIGOTimeGPS, f0: float,
    deltaT: float, sampleUnits: Unit, length: int,
) -> UINT8TimeSeries: ...

def CreateREAL4TimeSeries(
    name: str | None, epoch: LIGOTimeGPS, f0: float,
    deltaT: float, sampleUnits: Unit, length: int,
) -> REAL4TimeSeries: ...

def CreateREAL8TimeSeries(
    name: str | None, epoch: LIGOTimeGPS, f0: float,
    deltaT: float, sampleUnits: Unit, length: int,
) -> REAL8TimeSeries: ...

def CreateCOMPLEX8TimeSeries(
    name: str | None, epoch: LIGOTimeGPS, f0: float,
    deltaT: float, sampleUnits: Unit, length: int,
) -> COMPLEX8TimeSeries: ...

def CreateCOMPLEX16TimeSeries(
    name: str | None, epoch: LIGOTimeGPS, f0: float,
    deltaT: float, sampleUnits: Unit, length: int,
) -> COMPLEX16TimeSeries: ...

class _FrequencySeries(Generic[_TVector]):
    data: _TVector
    deltaF: float
    epoch: LIGOTimeGPS
    f0: float
    length: int
    name: str
    sampleUnits: Unit

class INT2FrequencySeries(_FrequencySeries[INT2Vector]): ...
class INT4FrequencySeries(_FrequencySeries[INT4Vector]): ...
class INT8FrequencySeries(_FrequencySeries[INT8Vector]): ...
class UINT2FrequencySeries(_FrequencySeries[UINT2Vector]): ...
class UINT4FrequencySeries(_FrequencySeries[UINT4Vector]): ...
class UINT8FrequencySeries(_FrequencySeries[UINT8Vector]): ...
class REAL4FrequencySeries(_FrequencySeries[REAL4Vector]): ...
class REAL8FrequencySeries(_FrequencySeries[REAL8Vector]): ...
class COMPLEX8FrequencySeries(_FrequencySeries[COMPLEX8Vector]): ...
class COMPLEX16FrequencySeries(_FrequencySeries[COMPLEX16Vector]): ...

def CreateINT2FrequencySeries(
    name: str | None, f0: float,
    deltaF: float, sampleUnits: Unit, length: int,
) -> INT2FrequencySeries: ...

def CreateINT4FrequencySeries(
    name: str | None, f0: float,
    deltaF: float, sampleUnits: Unit, length: int,
) -> INT4FrequencySeries: ...

def CreateINT8FrequencySeries(
    name: str | None, f0: float,
    deltaF: float, sampleUnits: Unit, length: int,
) -> INT8FrequencySeries: ...

def CreateUINT2FrequencySeries(
    name: str | None, f0: float,
    deltaF: float, sampleUnits: Unit, length: int,
) -> UINT2FrequencySeries: ...

def CreateUINT4FrequencySeries(
    name: str | None, f0: float,
    deltaF: float, sampleUnits: Unit, length: int,
) -> UINT4FrequencySeries: ...

def CreateUINT8FrequencySeries(
    name: str | None, f0: float,
    deltaF: float, sampleUnits: Unit, length: int,
) -> UINT8FrequencySeries: ...

def CreateREAL4FrequencySeries(
    name: str | None, f0: float,
    deltaF: float, sampleUnits: Unit, length: int,
) -> REAL4FrequencySeries: ...

def CreateREAL8FrequencySeries(
    name: str | None, f0: float,
    deltaF: float, sampleUnits: Unit, length: int,
) -> REAL8FrequencySeries: ...

def CreateCOMPLEX8FrequencySeries(
    name: str | None, f0: float,
    deltaF: float, sampleUnits: Unit, length: int,
) -> COMPLEX8FrequencySeries: ...

def CreateCOMPLEX16FrequencySeries(
    name: str | None, f0: float,
    deltaF: float, sampleUnits: Unit, length: int,
) -> COMPLEX16FrequencySeries: ...

# -- Window --------------------------

class _Window(Generic[_TVector]):
    data: _TVector
    sum: float
    sumofsquares: float

class REAL4Window(_Window[REAL4Vector]): ...
class REAL8Window(_Window[REAL8Vector]): ...

def CreateNamedREAL4Window(windowName: str, length: int) -> REAL4Window: ...
def CreateNamedREAL8Window(windowName: str, length: int) -> REAL8Window: ...

# -- FFTPlan -------------------------

class REAL4FFTPlan: ...
class REAL8FFTPlan: ...
class COMPLEX8FFTPlan: ...
class COMPLEX16FFTPlan: ...

def CreateREAL4FFTPlan(size: int, fwdflg: int, measurelvl: int) -> REAL4FFTPlan: ...
def CreateREAL8FFTPlan(size: int, fwdflg: int, measurelvl: int) -> REAL8FFTPlan: ...
def CreateCOMPLEX8FFTPlan(size: int, fwdflg: int, measurelvl: int) -> COMPLEX8FFTPlan: ...
def CreateCOMPLEX16FFTPlan(size: int, fwdflg: int, measurelvl: int) -> COMPLEX16FFTPlan: ...

# Average spectrum functions / types used by the spectral glue layer
class REAL8AverageSpectrumWelch: ...
class REAL8AverageSpectrumMedian: ...
class REAL8AverageSpectrumMedianMean: ...

# -- Sequence ------------------------

def CreateINT2Sequence(length: int) -> INT2Vector: ...
def CreateINT4Sequence(length: int) -> INT4Vector: ...
def CreateINT8Sequence(length: int) -> INT8Vector: ...
def CreateUINT2Sequence(length: int) -> UINT2Vector: ...
def CreateUINT4Sequence(length: int) -> UINT4Vector: ...
def CreateUINT8Sequence(length: int) -> UINT8Vector: ...
def CreateREAL8Sequence(length: int) -> REAL8Vector: ...
def CreateCOMPLEX8Sequence(length: int) -> COMPLEX8Vector: ...
def CreateCOMPLEX16Sequence(length: int) -> COMPLEX16Vector: ...

# -- Time ----------------------------

# LIGO time struct
class LIGOTimeGPS:
    gpsSeconds: int
    gpsNanoSeconds: int

    @overload
    def __init__(self) -> None: ...
    @overload
    def __init__(self, new: float | LIGOTimeGPS | str) -> None: ...
    @overload
    def __init__(self, seconds: int, nanoseconds: int) -> None: ...

    def __float__(self) -> float: ...
    def __int__(self) -> int: ...

    # equality: accept any object (Python semantics)
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...

    # ordering comparisons with other times or numeric offsets
    def __lt__(self, other: LIGOTimeGPS | float) -> bool: ...
    def __le__(self, other: LIGOTimeGPS | float) -> bool: ...
    def __gt__(self, other: LIGOTimeGPS | float) -> bool: ...
    def __ge__(self, other: LIGOTimeGPS | float) -> bool: ...

    # arithmetic: addition/subtraction with another time or numeric offset
    def __add__(self, other: LIGOTimeGPS | float) -> LIGOTimeGPS: ...
    def __radd__(self, other: float) -> LIGOTimeGPS: ...
    def __sub__(self, other: LIGOTimeGPS | float) -> LIGOTimeGPS: ...
    def __rsub__(self, other: float) -> LIGOTimeGPS: ...

    # unary
    def __neg__(self) -> LIGOTimeGPS: ...
    def __pos__(self) -> LIGOTimeGPS: ...

# Unit representation used by LAL. The real C struct exposes arrays of
# integers; expose as mutable sequences for simple indexed assignment.
class Unit:
    unitNumerator: list[int]
    unitDenominatorMinusOne: list[int]
    powerOfTen: int
    def __init__(self, spec: str | None = None) -> None: ...

DimensionlessUnit: Unit
MeterUnit: Unit
StrainUnit: Unit



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""GWF I/O backend discovery."""

from __future__ import annotations

import importlib
import os
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import (
        Iterable,
        Iterator,
    )
    from types import (
        FunctionType,
        ModuleType,
    )

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

__all__ = [
    "BACKENDS",
    "get_backend",
    "get_backend_function",
    "import_backend",
]

# Preferentially ordered list of supported GWF backends
BACKENDS = [
    "frameCPP",
    "LALFrame",
]


def _backend_candidates(
    backends: list[str] = BACKENDS,
) -> Iterator[str]:
    default = os.getenv("GWPY_FRAME_LIBRARY") or backends[0]
    yield from dict.fromkeys([default, *backends])


def get_backend(
    package: str = __package__,
    backends: list[str] = BACKENDS,
) -> str:
    """Return the preferred GWF backend.

    This can be configured via the ``GWPY_FRAME_LIBRARY`` environment
    variable, which can be set to the name of any of the interface modules
    defined under `gwpy.io.gwf` (or the module path given by ``package``):

    - ``"frameCPP"``
    - ``"FrameL"``
    - ``"LALFrame"``

    Otherwise that list is manually searched in the order given above.

    Parameters
    ----------
    package : `str`
        The package under which to attempt to install a GWF backend module.

    backends : `list` of `str`
        The list of backends to try, defaults to the list of backends
        implemented in :mod:`gwpy.io.gwf.backend`.

    Examples
    --------
    If the environment variable ``GWPY_FRAME_LIBRARY`` is set:

    >>> os.environ["GWPY_FRAME_LIBRARY"] = "FrameL"
    >>> from gwpy.io.gwf import get_backend
    >>> get_backend()
    'FrameL'

    Or, if you have |LDAStools.frameCPP|_ installed:

    >>> from gwpy.io.gwf import get_backend
    >>> get_backend()
    'frameCPP'

    Or, if you don't have |lalframe|_:

    >>> get_backend()
    'LALFrame'

    Otherwise:

    >>> get_backend()
    ImportError: no GWF API available, please install a third-party GWF
    library (frameCPP, FrameL, LALFrame) and try again
    """
    for lib in _backend_candidates(backends):
        try:
            import_backend(lib.lower(), package=package)
        except ImportError:
            continue
        else:
            return lib
    msg = (
        "no GWF API available, please install a third-party GWF "
        f"library ({', '.join(BACKENDS)}) and try again"
    )
    raise ImportError(msg)


def import_backend(
    library: str,
    package: str = __package__,
) -> ModuleType:
    """Utility method to import the relevant GWF I/O backend.

    This is just a wrapper around :meth:`importlib.import_module` with
    a slightly nicer error message.
    """
    # import the frame library here to have any ImportErrors occur early
    try:
        return importlib.import_module(f".{library.lower()}", package=package)
    except ImportError as exc:
        exc.args = (
            f"Cannot import {library} GWF backend: {exc}",
        )
        raise


def get_backend_function(
    name: str,
    backend: str | None = None,
    backends: list[str] = BACKENDS,
    package: str = __package__,
) -> FunctionType:
    """Return the backend implementation of the function of the given name.

    Parameters
    ----------
    name : `str`
        The name of the function to import.

    backend : `str`
        The specific backend to use.

    backends : `list` of `str`
        The list of backends to query, defaults to all known backends,
        ignored if ``backend`` is given.

    package : `str`
        The package from which to attempt to install the backend module,
        defaults to `gwpy.io.gwf`.
    """
    backends_to_try: Iterable[str]
    if backend:
        backends_to_try = [backend]
    else:
        backends_to_try = _backend_candidates(backends)
    for bck in backends_to_try:
        try:
            mod = import_backend(bck, package=package)
            return getattr(mod, name)
        except ImportError:  # module not installed
            if backend:
                raise
            continue
        except AttributeError as exc:  # function not implemented
            if backend:  # user asked for this backend specifically
                msg = f"GWF backend '{backend}' does not implement '{name}'"
                raise NotImplementedError(msg) from exc
            continue
    # Try and import any backend to see if anything is actually available
    get_backend(package=package, backends=backends)
    # Otherwise we know that the requested function is not implemented
    # in any available backend
    msg = f"no GWF backend module found that implements '{name}'"
    raise NotImplementedError(msg)



# Copyright (c) 2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for `Table.read` and `Table.write` with ``format='gwf'``."""

import pytest

from ...testing import utils
from ..filter import filter_table
from .utils import (
    TABLE_CLASSES,
    random_table,
)

# -- test config ---------------------

def pytest_generate_tests(metafunc):
    """Parametrize the ``table`` fixture for each valid table type."""
    if "table" in metafunc.fixturenames:
        metafunc.parametrize("table", TABLE_CLASSES, indirect=True)


@pytest.fixture
def table(request):
    """Create a random table with the right columns."""
    return random_table(
        length=100,
        names=["time", "frequency", "snr"],
        table_class=request.param,
    )


@pytest.fixture
def gwf_file(table, tmp_path):
    """Write the ``table`` to a (temporary) file and return the path.

    Enables read tests.
    """
    tmp = tmp_path / "table.gwf"
    table.write(tmp, "test")
    return tmp


# -- tests ---------------------------

@pytest.mark.requires("LDAStools.frameCPP")
def test_read_write_gwf(table):
    """Test `Table.read` and `Table.write` with ``format='gwf'``."""
    utils.test_read_write(
        table,
        "gwf",
        write_args=("test",),
        read_args=("test",),
        read_kw={
            "columns": table.colnames,
        },
        autoidentify=True,
        assert_equal=utils.assert_table_equal,
        assert_kw={
            "almost_equal": True,
        },
    )


@pytest.mark.requires("LDAStools.frameCPP")
def test_read_gwf_columns_where(table, gwf_file):
    """Test `Table.read(format='gwf')` with ``columns`` and ``where``."""
    # check where works
    t2 = type(table).read(
        gwf_file,
        "test",
        columns=["frequency", "snr"],
        where="frequency>500",
    )
    utils.assert_table_equal(
        t2,
        filter_table(table, "frequency>500")["frequency", "snr"],
    )



# Copyright (c) 2015-2020 Joseph Areeda
#               2020-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Coherence plots."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from ..plot import Plot
from ..plot.tex import label_to_latex
from .spectrum import SpectrumProduct

if TYPE_CHECKING:
    from argparse import (
        ArgumentParser,
        Namespace,
        _ArgumentGroup,
    )
    from logging import Logger
    from typing import ClassVar

    from matplotlib.legend import Legend

    from ..frequencyseries import FrequencySeries
    from ..plot import Axes
    from ..segments import Segment

__author__ = "Joseph Areeda <joseph.areeda@ligo.org>"

logger = logging.getLogger(__name__)


class CoherenceProduct(SpectrumProduct):
    """Plot coherence between two timeseries."""

    action: ClassVar[str] = "coherence"

    MIN_DATASETS: ClassVar[int] = 2

    def __init__(
        self,
        args: Namespace,
        logger: Logger = logger,
    ) -> None:
        """Create a new `Coherence` product."""
        super().__init__(args, logger=logger)
        self.ref_chan = self.args.ref or self.chan_list[0]
        # deal with channel type appendages
        self.ref_chan = self.ref_chan.split(",", 1)[0]

    @classmethod
    def arg_channels(cls, parser: ArgumentParser) -> _ArgumentGroup:
        """Set up channel arguments for this product."""
        group = super().arg_channels(parser)
        group.add_argument(
            "--ref",
            help="Reference channel against which others will be compared",
        )
        return group

    def _finalize_arguments(self, args: Namespace) -> None:
        if args.yscale is None:
            args.yscale = "linear"
        if args.yscale == "linear":
            if not args.ymin:
                args.ymin = 0
            if not args.ymax:
                args.ymax = 1.05
        return super()._finalize_arguments(args)

    def get_ylabel(self) -> str:
        """Text for y-axis label."""
        return "Coherence"

    def get_suptitle(self) -> str:
        """Start of default super title, first channel is appended to it."""
        return f"Coherence: {self.ref_chan}"

    def make_plot(self) -> Plot:
        """Generate the coherence plot from all time series."""
        args = self.args

        fftlength = float(args.secpfft)
        overlap = args.overlap
        self.logger.debug(
            "Calculating spectrum secpfft: %s, overlap: %s",
            fftlength,
            overlap,
        )
        if overlap is not None:
            overlap *= fftlength

        self.logger.debug("Reference channel: %s", self.ref_chan)

        # group data by segment
        groups: dict[Segment, dict[str, FrequencySeries]] = {}
        for series in self.timeseries:
            seg = series.span
            name = str(series.name or series.channel or "")
            try:
                groups[seg][name] = series
            except KeyError:
                groups[seg] = {}
                groups[seg][name] = series

        # -- plot

        plot = Plot(figsize=self.figsize, dpi=self.dpi)
        ax = plot.gca()
        self.spectra: list[FrequencySeries] = []

        # calculate coherence
        for group in groups.values():
            refts = group.pop(self.ref_chan)
            for name, series in group.items():
                coh = series.coherence(
                    refts,
                    fftlength=fftlength,
                    overlap=overlap,
                    window=args.window,
                )

                label = name
                if len(self.start_list) > 1:
                    label += f", {series.x0.value}"
                if self.usetex:
                    label = label_to_latex(label)

                ax.plot(coh, label=label)
                self.spectra.append(coh)

        if args.xscale == "log" and not args.xmin:
            args.xmin = 1/fftlength

        return plot

    def set_legend(self, ax: Axes | None = None) -> Legend | None:
        """Create a legend for this product."""
        leg = super().set_legend(ax=ax)
        if leg is not None:
            leg.set_title("Coherence with:")
        return leg



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2021 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for `gwpy.table.gravityspy`."""

from __future__ import annotations

from typing import TYPE_CHECKING
from urllib.parse import urlencode

import pytest
import requests

from ...testing.errors import pytest_skip_flaky_network
from .. import (
    GravitySpyTable,
    gravityspy as table_gravityspy,
)
from .test_table import TestEventTable as _TestEventTable

if TYPE_CHECKING:
    from pathlib import Path

    from requests_mock import Mocker as RequestMocker

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

TEST_ID = "8FHTgA8MEu"
TEST_IFO = "H1"


class TestGravitySpyTable(_TestEventTable[GravitySpyTable]):
    """Tests for `GravitySpyTable`."""

    TABLE: type[GravitySpyTable] = GravitySpyTable

    @classmethod
    def search(cls) -> GravitySpyTable:
        """Search for a known GravitySpy event."""
        return cls.TABLE.search(
            gravityspy_id=TEST_ID,
            howmany=1,
            ifos="H1L1",
            timeout=60,
        )

    @pytest_skip_flaky_network
    def test_search(self):
        """Test `GravitySpyTable.search`."""
        # run the search
        table = self.search()

        # validate the result
        assert len(table) == 1
        t = table[0]
        assert t["gravityspy_id"] == TEST_ID
        assert t["ifo"] == TEST_IFO
        assert t["ml_label"] == "Scratchy"

    def test_search_error(self, requests_mock: RequestMocker):
        """Test `GravitySpyTable.search` error handling.

        Mainly to make sure that an HTTP Error is raised instead
        of a JSONDecodeError.
        """
        url = (
            table_gravityspy.DEFAULT_HOST
            + table_gravityspy.SEARCH_PATH
            + table_gravityspy.SIMILARITY_SEARCH_PATH
            + "/?"
        )
        requests_mock.get(
            url + urlencode({
                "howmany": 10,
                "imageid": "abcde",
                "era": "event_time BETWEEN 1126400000 AND 1584057618",
                "ifo": "'H1', 'L1'",
                "database": "similarity_index_o3",
            }),
            text="<!DOCTYPE html><html></html>",
            status_code=200,
            headers={
                "Content-Type": "text/html; charset=utf-8",
            },
            complete_qs=True,
        )
        with pytest.raises(
            requests.HTTPError,
            match=r"please check the request parameters$",
        ):

            self.TABLE.search(gravityspy_id="abcde")

    @pytest_skip_flaky_network
    def test_download(self, tmp_path: Path):
        """Test `GravitySpyTable.download`."""
        table = self.search()
        table.download(download_path=tmp_path)
        assert (
            tmp_path
            / f"{TEST_IFO}_{TEST_ID}_spectrogram_0.5.png"
        ).is_file()



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Timezone utilities for GWOs."""

import datetime

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

# local time-zone for ground-based laser interferometers
TIMEZONE: dict[str, str] = {
    "C1": "US/Pacific",
    "G1": "Europe/Berlin",
    "H1": "US/Pacific",
    "I1": "Asia/Kolkata",
    "K1": "Japan",
    "L1": "US/Central",
    "V1": "Europe/Rome",
}


def get_timezone(ifo: str) -> str:
    """Return the timezone for the given interferometer prefix.

    Parameters
    ----------
    ifo : `str`
        Prefix of interferometer, e.g. ``'X1'``

    Returns
    -------
    timezone : `str`
        The name of the timezone for ``ifo``.

    Raises
    ------
    ValueError
        If ``ifo`` is not recognised.

    Examples
    --------
    >>> get_timezone("G1")
    'Europe/Berlin'
    """
    try:
        return TIMEZONE[ifo]
    except KeyError as exc:
        msg = f"Unrecognised ifo: '{ifo}'"
        raise ValueError(msg) from exc


def get_timezone_offset(ifo: str, dt: datetime.datetime | None = None) -> float:
    """Return the offset in seconds between UTC and the given interferometer.

    Parameters
    ----------
    ifo : `str`
        Prefix of interferometer, e.g. ``'X1'``

    dt : `datetime.datetime`, optional
        The time at which to calculate the offset.
        Default is `datetime.datetime.now`.

    Returns
    -------
    offset : `float`
        The offset in seconds between the timezone of the interferometer
        and UTC.
    """
    import pytz
    if dt is None:
        dt = datetime.datetime.now(tz=datetime.UTC)
    if dt.tzinfo is None:
        dt = dt.astimezone(datetime.UTC)
    tz = pytz.timezone(get_timezone(ifo))
    return dt.astimezone(tz).utcoffset().total_seconds()



# Copyright (c) 2023-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.timeseries.io.cache`."""

import pytest

from ...io.cache import write_cache
from ..io import cache as ts_io_cache


@pytest.fixture
def cache():
    """List of files over which to test sorting/sieving."""
    return [
        "/tmp/A-TEST-0-10.tmp",
        "/tmp/A-TEST-10-10.tmp",
        "/tmp/A-TEST-20-10.tmp",
        "/tmp/A-TEST-30-5.tmp",
        "/tmp/A-TEST-35-15.tmp",
    ]


@pytest.fixture
def cache_file(tmp_path, cache):
    """File version of `cache()`."""
    path = tmp_path / "cache.txt"
    write_cache(cache, path)
    return path


@pytest.mark.parametrize("source", ("cache", "cache_file"))
@pytest.mark.parametrize(("start", "end", "idx"), [
    # use everything in the cache
    (None, None, slice(None)),
    # use only GPS time '25' onwards, which is cache[2:]
    (25, None, slice(2, None)),
    # use only up to GPS time '25', which is cache[:3]
    (None, 25, slice(None, 3)),
    # use interval [10, 35), which needs cache[1:4]
    (10, 35, slice(1, 4)),
])
def test_preformat_cache(request, cache, source, start, end, idx):
    """Test that `gwpy.timeseries.io.cache.preformat_cache` works properly.

    Here `[start, end)` is a GPS segment, and `idx` the corresponding slice
    needed to restrict the cache object.

    Loops over a variety of input arguments, using `request` to dynamically
    loop over `cache` or `cache_file` as the input.
    """
    assert ts_io_cache.preformat_cache(
        request.getfixturevalue(source),  # cache or cache_file
        start=start,
        end=end,
    ) == cache[idx]



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""
.. sectionauthor:: Duncan Macleod <duncan.macleod@ligo.org>
.. currentmodule:: gwpy.timeseries

Generating a `SpectralVariance` histogram
#########################################

The most common visualisation of the spectral content of a data series is
via the power or amplitude spectral density calculations (PSD or ASD).
However, these typically involve averaging of the data over a period, which
can wash out transient noise (as is often desired).

The `SpectralVariance` histogram provide by `gwpy.frequencyseries` allows
us to look at the spectral sensitivity in a different manner, displaying
which frequencies sit at which amplitude *most* of the time, but also
highlighting excursions from normal behaviour.
"""

# %%
# To demonstate this, we can load some data from the LIGO Livingston
# interferometer around the time of the GW151226 gravitational wave detection:

from gwpy.timeseries import TimeSeries
llo = TimeSeries.get("L1", 1135136228, 1135140324)

# %%
# We can then call the :meth:`~gwpy.timeseries.TimeSeries.spectral_variance`
# method of the ``llo`` `~gwpy.timeseries.TimeSeries` by calculating an ASD
# every 5 seconds and counting the amount of time each frequency bin spends
# at each ASD value:

variance = llo.spectral_variance(
    5,
    fftlength=2,
    overlap=1,
    log=True,
    low=1e-24,
    high=1e-19,
    nbins=100,
)

# %%
# We can then :meth:`~SpectralVariance.plot` the `SpectralVariance`

plot = variance.plot(yscale="log", norm="log", vmin=.5, cmap="plasma")
ax = plot.gca()
ax.grid()
ax.set_xlim(20, 1500)
ax.set_ylim(1e-24, 1e-20)
ax.set_xlabel("Frequency [Hz]")
ax.set_ylabel(r"[strain/$\sqrt{\mathrm{Hz}}$]")
ax.set_title("LIGO-Livingston sensitivity variance")
plot.show()

# %%
# From this we see that in general the sensitivity varies a few parts in
# 10:sup:`-23`, while many of the lines (narrow-band peaks in the spectrum)
# are much more stationary.



# Copyright (c) 2021-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Error handling for the GWpy test suite."""

from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING

import pytest
import requests.exceptions
from igwn_segments import segment

from ..io.remote import NETWORK_ERROR

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import (
        ParamSpec,
        TypeVar,
    )

    P = ParamSpec("P")
    R = TypeVar("R")

# -- Network/HTTP --------------------

SERVER_ERROR = segment(500, 600)

# attempt to also catch errors from pytest_socket,
# this should enable the test suite to run on machines that
# don't allow network access, e.g. debian build machines
try:
    from pytest_socket import (
        SocketBlockedError,
        SocketConnectBlockedError,
    )
except ModuleNotFoundError:  # pragma: no cover
    # pytest-socket not installed
    pass
except ImportError as exc:  # pragma: no cover
    # pytest-socket installed but errors not found,
    # print a warning to tell the devs to update this module
    import warnings
    warnings.warn(
        f"failed to import exception types from pytest_socket: {exc}",
        stacklevel=2,
    )
else:
    NETWORK_ERROR = (
        *NETWORK_ERROR,
        SocketBlockedError,
        SocketConnectBlockedError,
    )


pytest_rerun_flaky_httperror = pytest.mark.flaky(
    reruns=2,
    reruns_delay=5,
    only_rerun=["HTTPError"],
)


def pytest_skip_network_error(
    func: Callable[P, R],
) -> Callable[P, R]:
    """Execute `func` but skip if it raises one of the network exceptions."""
    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        try:
            return func(*args, **kwargs)
        except* NETWORK_ERROR as exc:  # pragma: no cover
            if not exc.message and len(exc.exceptions) == 1:
                exc = exc.exceptions[0]  # type: ignore[assignment]
            pytest.skip(str(exc))
        except* requests.exceptions.HTTPError as exc:  # pragma: no cover
            for error in exc.exceptions:
                if (
                    isinstance(error, requests.exceptions.HTTPError)
                    and error.response.status_code not in SERVER_ERROR
                ):
                    raise
            pytest.skip(str(exc.exceptions[0]))

    return wrapper


def pytest_skip_flaky_network(
    func: Callable,
) -> Callable:
    """Decorate test ``func`` with all necessary network decorators.

    A test decorated with this decorator will attempt behave as follows

    - if an `HTTPError` is encountered, retry a couple of times with
      a 5-second delay, then 'skip' if the response is a 5xx code,
      otherwise fail,
    - if any non-HTTP network-related error is encountered, skip
      the test.
    """
    for dec in (
        pytest_skip_network_error,
        pytest_rerun_flaky_httperror,
    ):
        func = dec(func)
    return func



# Copyright (c) 2024-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unified I/O read/write for Segment objects."""

from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

from ..io.registry import (
    UnifiedRead,
    UnifiedWrite,
)

if TYPE_CHECKING:
    from collections.abc import Sequence

    from ..io.utils import NamedReadable
    from . import (
        DataQualityDict,
        DataQualityFlag,
        SegmentList,
    )


class SegmentListRead(UnifiedRead):
    """Read segments from file into a `SegmentList`.

    Parameters
    ----------
    filename : `str`
        Path of file to read.

    format : `str`, optional
        Source format identifier. If not given, the format will be
        detected if possible. See below for list of acceptable
        formats.

    coalesce : `bool`, optional
        If `True`, coalesce the segment list before returning,
        otherwise return exactly as contained in file(s).

    kwargs
        Other keyword arguments depend on the format,
        see `SegmentList.read.help(format=<format>)` for details, or
        see the online documentation (:ref:`gwpy-segments-io`).

    Returns
    -------
    segmentlist : `SegmentList`
        `SegmentList` active and known segments read from file.

    Raises
    ------
    IndexError
        if ``source`` is an empty list

    Notes
    -----"""

    def merge(  # type: ignore[override]
        self,
        items: Sequence[SegmentList],
        *,
        coalesce: bool = False,
    ) -> SegmentList:
        """Combine the `SegmentList` from each file into a single object."""
        out = self._cls(seg for seglist in items for seg in seglist)
        if coalesce:
            return out.coalesce()
        return out


class SegmentListWrite(UnifiedWrite):
    """Write this `SegmentList` to a file.

    Arguments and keywords depend on the output format, see the
    online documentation for full details for each format.

    Parameters
    ----------
    target : `str`
        Output filename.

    Notes
    -----"""


class DataQualityFlagRead(UnifiedRead):
    """Read segments from file into a `DataQualityFlag`.

    Parameters
    ----------
    filename : `str`
        Path of file to read.

    name : `str`, optional
        Name of flag to read from file, otherwise read all segments.

    format : `str`, optional
        Source format identifier. If not given, the format will be
        detected if possible. See below for list of acceptable
        formats.

    coltype : `type`, optional, default: `float`
        Datatype to force for segment times, only valid for
        ``format='segwizard'``.

    strict : `bool`, optional, default: `True`
        Require segment start and stop times match printed duration,
        only valid for ``format='segwizard'``.

    coalesce : `bool`, optional
        If `True` coalesce the all segment lists before returning,
        otherwise return exactly as contained in file(s).

    parallel : `int`, optional, default: 1
        Number of threads to use for parallel reading of multiple files.

    verbose : `bool`, optional, default: `False`
        Print a progress bar showing read status.

    Returns
    -------
    dqflag : `DataQualityFlag`
        formatted `DataQualityFlag` containing the active and known
        segments read from file.

    Raises
    ------
    IndexError
        If ``source`` is an empty list.

    Notes
    -----"""

    def merge(  # type: ignore[override]
        self,
        items: Sequence[DataQualityFlag],
        *,
        coalesce: bool = False,
    ) -> DataQualityFlag:
        """Combine `DataQualityFlag` from each file into a single object."""
        itrtr = iter(items)
        out = next(itrtr)
        for flag in itrtr:
            out.known += flag.known
            out.active += flag.active
        if coalesce:
            return out.coalesce()
        return out

    def __call__(
        self,
        source: NamedReadable | list[NamedReadable],
        *args,
        coalesce: bool = False,
        **kwargs,
    ) -> DataQualityFlag:
        """Read data into a `DataQualityFlag`."""
        if "flag" in kwargs:  # pragma: no cover
            warnings.warn(
                "'flag' keyword was renamed 'name', "
                "this warning will result in an error in the future",
                DeprecationWarning,
                stacklevel=2,
            )
            kwargs.setdefault("name", kwargs.pop("flag"))

        return super().__call__(
            source,
            *args,
            coalesce=coalesce,
            **kwargs,
        )


class DataQualityFlagWrite(UnifiedWrite):
    """Write this `DataQualityFlag` to file.

    Notes
    -----"""


class DataQualityDictRead(UnifiedRead):
    """Read segments from file into a `DataQualityDict`.

    Parameters
    ----------
    source : `str`
        Path of file to read

    format : `str`, optional
        Source format identifier. If not given, the format will be
        detected if possible. See below for list of acceptable
        formats.

    names : `list`, optional, default: read all names found
        List of names to read, by default all names are read separately.

    coalesce : `bool`, optional
        If `True` coalesce the all segment lists before returning,
        otherwise return exactly as contained in file(s).

    parallel : `int`, optional, default: 1
        Number of threads to use for parallel reading of multiple files.

    verbose : `bool`, optional, default: `False`
        Print a progress bar showing read status.

    Returns
    -------
    flagdict : `DataQualityDict`
        A new `DataQualityDict` of `DataQualityFlag` entries with
        ``active`` and ``known`` segments seeded from the XML tables
        in the given file.

    Notes
    -----"""

    def merge(  # type: ignore[override]
        self,
        inputs: Sequence[DataQualityDict],
        *,
        names: list[str] | None = None,
        on_missing: str = "error",
        coalesce: bool = False,
    ) -> DataQualityDict:
        """Combine a list of `DataQualityDict` objects into one `DataQualityDict`.

        Must be given at least one `DataQualityDict`.
        """
        # check all names are contained
        required = set(names or [])
        found = {name for dqdict in inputs for name in dqdict}
        for name in required - found:  # validate all names are found once
            msg = f"'{name}' not found in any input file"
            if on_missing == "ignore":
                continue
            if on_missing == "warn":
                warnings.warn(msg, stacklevel=2)
                continue
            raise ValueError(msg)

        # combine flags
        out = self._cls()
        for dqdict in inputs:
            for flag in dqdict:
                if flag in out:
                    out[flag].known.extend(dqdict[flag].known)
                    out[flag].active.extend(dqdict[flag].active)
                else:
                    out[flag] = dqdict[flag]
        if coalesce:
            return out.coalesce()
        return out

    def __call__(
        self,
        *args,
        names: list[str] | None = None,
        on_missing: str = "error",
        coalesce: bool = False,
        **kwargs,
    ):
        return super().__call__(
            *args,
            names=names,
            on_missing=on_missing,
            coalesce=coalesce,
            **kwargs,
        )


class DataQualityDictWrite(UnifiedWrite):
    """Write this `DataQualityDict` to file.

    Notes
    -----"""



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>

"""
.. sectionauthor:: Duncan Macleod <duncan.macleod@ligo.org>
.. currentmodule:: gwpy.frequencyseries

Calculate and plot a `FrequencySeries`
######################################

One of the principal means of estimating the sensitivity of a
gravitational-wave detector is to esimate it's amplitude spectral density
(ASD).
The ASD is a measurement of how a signal's amplitude varies across
different frequencies.

The ASD can be estimated directly from a `~gwpy.timeseries.TimeSeries`
using the :meth:`~gwpy.timeseries.TimeSeries.asd` method.
"""

# %%
# Data access
# -----------
# For this example we choose to estimate the ASD around |GW200115|_,
# one of the first observations of a neutron star-black hole binary.
# We can use the :doc:`gwosc <gwosc:index>` Python package to query
# for the relevant GPS time:

from gwosc.datasets import event_gps

gps = event_gps("GW200115")

# %%
# In order to generate a `FrequencySeries` we need to import the
# `~gwpy.timeseries.TimeSeries` and use
# :meth:`~gwpy.timeseries.TimeSeries.fetch_open_data` to download the strain
# records:

from gwpy.timeseries import TimeSeries

lho = TimeSeries.get("H1", gps - 16, gps + 16)
llo = TimeSeries.get("L1", gps - 16, gps + 16)

# %%
# Calculate the ASDs
# ------------------
# We can then call the :meth:`~gwpy.timeseries.TimeSeries.asd` method to
# calculated the amplitude spectral density for each
# `~gwpy.timeseries.TimeSeries`:

lhoasd = lho.asd(4, 2)
lloasd = llo.asd(4, 2)

# %%
# Visualisation
# -------------
# We can then :meth:`~FrequencySeries.plot` the spectra using the 'standard'
# colour scheme:

plot = lhoasd.plot(label="LIGO-Hanford", color="gwpy:ligo-hanford")
ax = plot.gca()
ax.plot(lloasd, label="LIGO-Livingston", color="gwpy:ligo-livingston")
ax.set_xlim(16, 1600)
ax.set_ylim(1e-24, 1e-21)
ax.set_ylabel(r"Strain ASD [1/$\sqrt{\mathrm{Hz}}]$")
ax.legend(frameon=False, bbox_to_anchor=(1., 1.), loc="lower right", ncol=2)
plot.show()



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read and write HDF5 files in GWOSC format.

For more details, see :ref:`gwpy-table-io`.
"""

from __future__ import annotations

import logging
import re
import warnings
from concurrent.futures import (
    ThreadPoolExecutor,
    as_completed,
)
from itertools import product
from math import ceil
from multiprocessing import cpu_count
from pathlib import Path
from typing import TYPE_CHECKING
from urllib.parse import urlparse

from astropy.units import Quantity
from gwosc.api import DEFAULT_URL as DEFAULT_GWOSC_URL
from gwosc.locate import get_urls

from ...detector import Channel
from ...detector.units import parse_unit
from ...io import (
    gwf as io_gwf,
    hdf5 as io_hdf5,
)
from ...io.cache import (
    file_segment,
    sieve as sieve_cache,
)
from ...io.remote import open_remote_file
from ...io.utils import file_path
from ...segments import Segment
from ...time import to_gps
from .. import (
    StateVector,
    TimeSeries,
    TimeSeriesBase,
    TimeSeriesBaseDict,
)

# Module logger
logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from collections.abc import (
        Collection,
        Iterable,
    )
    from typing import (
        IO,
        Any,
        TypeVar,
    )

    import h5py

    from ...time import SupportsToGps

    T = TypeVar("T", bound=TimeSeriesBase)


KHZ_4 = 4096
KHZ_16 = 16384

DQMASK_CHANNEL_REGEX = re.compile(r"\A[A-Z]\d:(GW|L)OSC-.*DQMASK\Z")
STRAIN_CHANNEL_REGEX = re.compile(r"\A[A-Z]\d:(GW|L)OSC-.*STRAIN\Z")

GWOSC_HDF5_FORMAT = "hdf5.gwosc"
GWOSC_LOCATE_KWARGS = (
    "sample_rate",
    "version",
    "host",
    "format",
    "dataset",
)

# Keyword arguments that we may inherit from TimeSeries.get
# that are unsafe to use in to fetch_gwosc_data
IGNORE_GET_KWARGS = {
    "frametype",
    "frametype_match",
    "urltype",
}

NUM_THREADS = min(4, cpu_count() or 1)


# -- utilities -----------------------

def _is_gwosc_channel(name: str | Channel) -> bool:
    """Check if a channel name looks like a GWOSC channel."""
    return ":GWOSC-" in str(name)


def _all_gwosc_channels(channels: Iterable[str | Channel]) -> bool:
    """Check if all channel names look like GWOSC channels."""
    return all(map(_is_gwosc_channel, channels))


def _any_gwosc_channels(channels: Iterable[str | Channel]) -> bool:
    """Check if any channel names look like GWOSC channels."""
    return any(map(_is_gwosc_channel, channels))


def _get_file_extension(url: str) -> str:
    """Get the file extension from a URL, handling compressed files."""
    if url.endswith(".gz"):
        path = Path(url[:-3])
    else:
        path = Path(url)
    return path.suffix


def _default_format(ext: str) -> str | None:
    """Set format in kwargs based on file extension."""
    if ext == ".hdf5":
        return GWOSC_HDF5_FORMAT
    if ext == ".gwf":
        return "gwf"
    return None


def _gwf_channel(
    source: IO | str,
    series_class: type[TimeSeriesBase] = TimeSeries,
) -> str:
    """Find the right channel name for a GWOSC GWF file."""
    channels = list(io_gwf.iter_channel_names(file_path(source)))
    if issubclass(series_class, StateVector):
        regex = DQMASK_CHANNEL_REGEX
    else:
        regex = STRAIN_CHANNEL_REGEX
    found, = list(filter(regex.match, channels))
    logger.debug("Using channel %r", found)
    return found


def _parse_bits_from_gwf_unit(series: StateVector) -> None:
    """Parse bit definitions from GWF unit string."""
    try:
        bits: dict[int, str | None] = {}
        for bit in str(series.unit).split():
            a, b = bit.split(":", 1)
            bits[int(a)] = b
        series.bits = bits
        series.override_unit("")
    except (TypeError, ValueError):
        # Don't care, bad GWOSC format
        pass


def _fetch_gwosc_data_file(
    url: str,
    *args: str | None,
    series_class: type[TimeSeriesBase] = TimeSeries,
    cache: bool | None = None,
    timeout: float | None = None,
    format: str | None = None,  # noqa: A002
    **kwargs,
) -> TimeSeriesBase:
    """Fetch a single GWOSC file and return a `Series`."""
    # Match file format
    ext = _get_file_extension(url)
    fmt = format or _default_format(ext)

    logger.debug("Downloading GWOSC data from %s", url)
    with open_remote_file(
        url,
        cache=cache,
        remote_timeout=timeout,
    ) as rem:
        # Get channel for GWF if not given
        if ext == ".gwf" and (not args or args[0] is None):
            args = (_gwf_channel(rem, series_class),)

        logger.debug("Reading %s", url.rsplit("/", maxsplit=1)[-1])

        try:
            series = series_class.read(rem, *args, format=fmt, **kwargs)
        except Exception as exc:
            exc.args = (f"Failed to read GWOSC data from {url!r}: {exc}",)
            raise
        else:
            # Parse bits from unit in GWF
            if ext == ".gwf" and isinstance(series, StateVector):
                _parse_bits_from_gwf_unit(series)
            return series


def _overlapping(files: Iterable[str]) -> bool:
    """Quick method to see if a file list contains overlapping files."""
    segments: set[Segment] = set()
    for path in files:
        seg = file_segment(path)
        for s in segments:
            if seg.intersects(s):
                return True
        segments.add(seg)
    return False


def _name_from_gwosc_hdf5(
    h5ds: h5py.Dataset,
    name: str = "GWOSCmeta",
) -> str:
    """Construct a channel-like name for a dataset in a GWOSC HDF5 file.

    New files (starting at O2 circa 2016-2017) contain a dataset called
    ``GWOSCmeta`` alongside the main dataset that stores the GWF channel
    (without the ifo name) so we use this to reconstruct the GWF channel.
    It works for strain, DQ and injections.

    For old files, we return the name of the dataset prefixed with the
    detector prefix from the ``/meta/Detector`` dataset.
    """
    h5g = h5ds.parent

    # New files store the channel name in GWOSCmeta
    try:
        meta_ds = h5g[name]
    except KeyError:  # no GWOSCmeta
        channel = h5ds.name.rsplit("/", maxsplit=1)[-1]
    else:
        channel = meta_ds[()].decode("utf-8")
    # We can then find the detector prefix
    ifo_ds = io_hdf5.find_dataset(h5g.file, "meta/Detector")
    ifo = ifo_ds[()].decode("utf-8")
    return f"{ifo}:{channel}"


# -- remote data access (the main event)

def fetch_gwosc_data(
    detector: str,
    start: SupportsToGps,
    end: SupportsToGps,
    dataset: str | None = None,
    version: int | None = None,
    sample_rate: int = 4096,
    format: str = "hdf5",
    host: str = DEFAULT_GWOSC_URL,
    series_class: type[T] = TimeSeries,
    *,
    verbose: bool | None = None,
    **kwargs,
) -> T:
    """Fetch open-access data from GWOSC.

    Parameters
    ----------
    detector : `str`
        The two-character prefix of the IFO in which you are interested,
        e.g. `'L1'`.

    start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS start time of required data, defaults to start of data found;
        any input parseable by `~gwpy.time.to_gps` is fine.

    end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS end time of required data, defaults to end of data found;
        any input parseable by `~gwpy.time.to_gps` is fine.

    sample_rate : `float`, optional,
        The sample rate of desired data; most data are stored
        by GWOSC at 4096 Hz, however there may be event-related
        data releases with a 16384 Hz rate, default: `4096`.

    version : `int`, optional
        Version of files to download, defaults to highest discovered
        version.

    format : `str`, optional
        The data format to download and parse, default: ``'h5py'``

        - ``'hdf5'``
        - ``'gwf'`` - requires `lalframe`

    host : `str`, optional
        HTTP host name of GWOSC server to access.

    cache : `bool`, optional
        Save/read a local copy of the remote URL, default: `False`;
        useful if the same remote data are to be accessed multiple times.
        Set ``GWPY_CACHE=1`` in the environment to auto-cache.

    timeout : `float`, optional
        The time to wait for a response from the GWOSC server.

    verbose : `bool`, optional
        This argument is deprecated and will be removed in a future release.
        Use DEBUG-level logging instead, see :ref:`gwpy-logging`.

    kwargs
        Any other keyword arguments are passed to the `TimeSeries.read`
        method that parses the file that was downloaded.

    Examples
    --------
    >>> from gwpy.timeseries import (TimeSeries, StateVector)
    >>> print(TimeSeries.get('H1', 1126259446, 1126259478))
    TimeSeries([  2.17704028e-19,  2.08763900e-19,  2.39681183e-19,
                ...,   3.55365541e-20,  6.33533516e-20,
                  7.58121195e-20]
               unit: Unit(dimensionless),
               t0: 1126259446.0 s,
               dt: 0.000244140625 s,
               name: Strain,
               channel: None)
    >>> print(StateVector.get('H1', 1126259446, 1126259478))
    StateVector([127,127,127,127,127,127,127,127,127,127,127,127,
                 127,127,127,127,127,127,127,127,127,127,127,127,
                 127,127,127,127,127,127,127,127]
                unit: Unit(dimensionless),
                t0: 1126259446.0 s,
                dt: 1.0 s,
                name: quality/simple,
                channel: None,
                bits: Bits(0: data present
                           1: passes cbc CAT1 test
                           2: passes cbc CAT2 test
                           3: passes cbc CAT3 test
                           4: passes burst CAT1 test
                           5: passes burst CAT2 test
                           6: passes burst CAT3 test,
                           channel=None,
                           epoch=1126259446.0))

    For the `StateVector`, the naming of the bits will be
    ``format``-dependent, because they are recorded differently by GWOSC
    in different formats.

    Notes
    -----
    `StateVector` data are not available in ``txt.gz`` format.
    """
    if verbose is not None:
        warnings.warn(
            "The 'verbose' argument is deprecated and will be removed in a future "
            "release, please consider using DEBUG-level logging instead.",
            DeprecationWarning,
            stacklevel=2,
        )

    # strip out arguments for other formats
    for key in IGNORE_GET_KWARGS:
        kwargs.pop(key, None)

    # format arguments
    start = to_gps(start)
    end = to_gps(end)
    span: Segment[float] = Segment(start, end)

    # find URLs (requires python-gwosc)
    sample_rate = int(Quantity(sample_rate, "Hz").value)
    urls = get_urls(
        detector,
        int(start),
        ceil(end),
        dataset=dataset,
        version=version,
        sample_rate=sample_rate,
        format=format,
        host=host,
    )
    cache = sieve_cache(urls, segment=span)

    # if event dataset, pick shortest file that covers the request
    # -- this is a bit hacky, and presumes that only an event dataset
    # -- would be produced with overlapping files.
    # -- This should probably be improved to use dataset information
    if len(cache) and _overlapping(cache):
        cache.sort(key=lambda x: abs(file_segment(x)))
        for url in cache:
            a, b = file_segment(url)
            if a <= start and b >= end:
                cache = [url]
                break
    logger.debug(
        "Fetched %d URLs from %s for [%s .. %s])",
        len(cache),
        urlparse(cache[0]).netloc,
        start,
        ceil(end),
    )

    is_gwf = cache[0].endswith(".gwf")
    args: tuple[str | Channel | None, ...]
    if is_gwf and cache:
        args = (kwargs.pop("channel", None),)
    else:
        args = ()

    # read data
    out = None
    kwargs["series_class"] = series_class
    for url in cache:
        keep = file_segment(url) & span
        kwargs["start"], kwargs["end"] = keep
        new = _fetch_gwosc_data_file(url, *args, **kwargs)
        if is_gwf and (not args or args[0] is None):
            args = (new.name,)
        if out is None:
            out = new.copy()
        else:
            out.append(new, resize=True)
    if out is None:
        msg = f"No data found for {detector} in [{start} .. {end})"
        raise ValueError(msg)
    return out


def fetch_dict(
    detectors: Collection[str | Channel],
    start: SupportsToGps,
    end: SupportsToGps,
    dataset: str | None = None,
    version: int | None = None,
    sample_rate: int = 4096,
    format: str = "hdf5",
    host: str = DEFAULT_GWOSC_URL,
    parallel: int = NUM_THREADS,
    series_class: type[T] = TimeSeries,
    **kwargs,
) -> dict[str | Channel, T]:
    """Fetch open-access data from GWOSC for multiple detectors.

    Parameters
    ----------
    detectors : `list` of `str`
        List of two-character prefices of the IFOs in which you
        are interested, e.g. `['H1', 'L1']`.

    start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
        GPS start time of required data,
        any input parseable by `~gwpy.time.to_gps` is fine.

    end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
        GPS end time of required data,
        any input parseable by `~gwpy.time.to_gps` is fine.

    sample_rate : `float`, `Quantity`,
        The sample rate (Hertz) of desired data; most data are stored
        by GWOSC at 4096 Hz, however there may be event-related
        data releases with a 16384 Hz rate.

    version : `int`
        Version of files to download, defaults to highest discovered
        version.

    format : `str`
        The data format to download and parse.
        One of

        "hdf5"
            HDF5 data files, read using `h5py`.

        "gwf"
            Gravitational-Wave Frame files, requires |LDAStools.frameCPP|_.

    host : `str`
        Host name of GWOSC server to access.

    cache : `bool`
        Save/read a local copy of the remote URL, default: `False`;
        useful if the same remote data are to be accessed multiple times.
        Set `GWPY_CACHE=1` in the environment to auto-cache.

    parallel : `int`
        Number of parallel threads to use when downloading data for
        multiple detectors. Default is ``1``.

    kwargs
        Any other keyword arguments are passed to the `TimeSeries.read`
        method that parses the file that was downloaded.

    See Also
    --------
    TimeSeries.fetch_open_data
        For more examples.

    TimeSeries.read
        For details of how files are read.

    Examples
    --------
    >>> from gwpy.timeseries import TimeSeriesDict
    >>> print(TimeSeriesDict.get(['H1', 'L1'], 1126259446, 1126259478))
    TimeSeriesDict({'H1': <TimeSeries([2.17704028e-19, 2.08763900e-19, 2.39681183e-19, ...,
                 3.55365541e-20, 6.33533516e-20, 7.58121195e-20]
                unit=Unit(dimensionless),
                t0=<Quantity 1.12625945e+09 s>,
                dt=<Quantity 0.00024414 s>,
                name='Strain',
                channel=None)>, 'L1': <TimeSeries([-1.04289994e-18, -1.03586274e-18, -9.89322445e-19,
                 ..., -1.01767748e-18, -9.82876816e-19,
                 -9.59276974e-19]
                unit=Unit(dimensionless),
                t0=<Quantity 1.12625945e+09 s>,
                dt=<Quantity 0.00024414 s>,
                name='Strain',
                channel=None)>})
    """  # noqa: E501
    names = {str(x).split(":", maxsplit=1)[0]: x for x in detectors}
    parallel = min(len(detectors), parallel or 1)

    def _fod(ifo: str) -> tuple[str, TimeSeriesBase]:
        """Fetch data for a single detector."""
        return ifo, fetch_gwosc_data(
            ifo,
            start,
            end,
            dataset=dataset,
            version=version,
            sample_rate=sample_rate,
            format=format,
            host=host,
            parallel=1,
            series_class=series_class,
            **kwargs,
        )

    # fetch all data in a thread pool
    out = series_class.DictClass()
    with ThreadPoolExecutor(max_workers=parallel) as pool:
        futures = [pool.submit(_fod, ifo) for ifo in names]
        for future in as_completed(futures):
            ifo, data = future.result()
            out[name := names[ifo]] = data
            logger.debug("Fetched data for %s", name)

    return out


# -- I/O -----------------------------

@io_hdf5.with_read_hdf5
def read_gwosc_hdf5(
    h5f: str | h5py.HLObject,
    path: str = "strain/Strain",
    start: SupportsToGps | None = None,
    end: SupportsToGps | None = None,
    *,
    copy: bool = False,
) -> TimeSeries:
    """Read a `TimeSeries` from a GWOSC-format HDF file.

    Parameters
    ----------
    h5f : `str`, `h5py.HLObject`
        path of HDF5 file, or open `H5File`

    path : `str`
        name of HDF5 dataset to read.

    start : `Time`, `~gwpy.time.LIGOTimeGPS`, optional
        start GPS time of desired data

    end : `Time`, `~gwpy.time.LIGOTimeGPS`, optional
        end GPS time of desired data

    copy : `bool`, default: `False`
        create a fresh-memory copy of the underlying array

    Returns
    -------
    data : `~gwpy.timeseries.TimeSeries`
        a new `TimeSeries` containing the data read from disk
    """
    dataset = io_hdf5.find_dataset(h5f, path)
    # read data
    nddata = dataset[()]
    # read metadata
    xunit = parse_unit(dataset.attrs["Xunits"])
    epoch = dataset.attrs["Xstart"]
    dt = Quantity(dataset.attrs["Xspacing"], xunit)
    unit = dataset.attrs["Yunits"]
    # build and return
    name = _name_from_gwosc_hdf5(dataset)
    return TimeSeries(
        nddata,
        epoch=epoch,
        sample_rate=(1/dt).to("Hertz"),
        unit=unit,
        name=name,
        copy=copy,
    ).crop(start=start, end=end)


@io_hdf5.with_read_hdf5
def read_gwosc_hdf5_state(
    f: str | h5py.HLObject,
    path: str = "quality/simple",
    start: SupportsToGps | None = None,
    end: SupportsToGps | None = None,
    *,
    copy: bool = False,
    value_dataset: str = "DQmask",
    bits_dataset: str = "DQDescriptions",
) -> StateVector:
    """Read a `StateVector` from a GWOSC-format HDF file.

    Parameters
    ----------
    f : `str`, `h5py.HLObject`
        path of HDF5 file, or open `H5File`

    path : `str`
        path of HDF5 datasets to read (will be used as name of the dataset).

    start : `Time`, `~gwpy.time.LIGOTimeGPS`, optional
        start GPS time of desired data

    end : `Time`, `~gwpy.time.LIGOTimeGPS`, optional
        end GPS time of desired data

    copy : `bool`, default: `False`
        create a fresh-memory copy of the underlying array

    value_dataset : `str`
        HDF5 dataset where to read the statevector values

    bits_dataset : `str`
        HDF5 dataset where to read the definition of each bits

    Returns
    -------
    data : `~gwpy.timeseries.StateVector`
        a new `StateVector` containing the data read from disk
    """
    # find data
    bits_ds = io_hdf5.find_dataset(f, f"{path}/{value_dataset}")
    def_ds = io_hdf5.find_dataset(f, f"{path}/{bits_dataset}")
    # read data
    bits = bits_ds[()]
    bit_def = [bytes.decode(bytes(b), "utf-8") for b in def_ds[()]]
    # read metadata
    epoch = bits_ds.attrs["Xstart"]
    try:
        dt = bits_ds.attrs["Xspacing"]
    except KeyError:
        dt = Quantity(1, "s")
    else:
        xunit = parse_unit(bits_ds.attrs["Xunits"])
        dt = Quantity(dt, xunit)
    # Name
    name = _name_from_gwosc_hdf5(bits_ds)
    return StateVector(
        bits,
        bits=bit_def,
        t0=epoch,
        name=name,
        dx=dt,
        copy=copy,
    ).crop(start=start, end=end)


TimeSeries.read.registry.register_reader(
    GWOSC_HDF5_FORMAT,
    TimeSeries,
    read_gwosc_hdf5,
)
StateVector.read.registry.register_reader(
    GWOSC_HDF5_FORMAT,
    StateVector,
    read_gwosc_hdf5_state,
)


# -- TimeSeries.get integration ------

def _as_tuple(
    value: str | int | Iterable[str | int] | None,
    default: Iterable[str | int],
) -> tuple[str | int, ...]:
    if value is None:
        return tuple(default)
    if isinstance(value, str | int):
        return (value,)
    return tuple(value)


def identify_gwosc_sources(
    origin: str,
    channels: str | Channel | Iterable[str | Channel],
    *args,  # noqa: ARG001, ANN002
    host: str | None = None,
    format: str | Iterable[str] | None = None,  # noqa: A002
    sample_rate: int | Iterable[int] | None = None,
    **kwargs,  # noqa: ARG001
) -> Iterable[dict[str, Any]] | None:
    """Identify GWOSC sources for these arguments."""
    if origin != "get":
        return None

    sources: list[dict[str, object]] = []

    # If host was given, and looks like nds2/gwdatafind, stop
    if str(host).startswith((
        "datafind",
        "gwdatafind",
        "nds",
    )):
        return sources

    channels = (channels,) if isinstance(channels, str | Channel) else channels
    names = {str(c) for c in channels}

    # Set priority (and format) based on channel names
    priority = 10
    if format is None and _all_gwosc_channels(names):
        format = "gwf"  # noqa: A001
        priority = 5  # user mentioned GWOSC, use it if we can
    elif any(":" in str(c) for c in channels):
        priority = 1000  # GWOSC almost certainly won't be able to help

    hosts = _as_tuple(host, (DEFAULT_GWOSC_URL,))
    formats = _as_tuple(format, ("hdf5", "gwf"))

    # Check if sample rate can be inferred from channel names
    if sample_rate is None:
        if all("4KHZ" in n for n in names):
            sample_rate = KHZ_4
        elif all("16KHZ" in n for n in names):
            sample_rate = KHZ_16
    sample_rates = _as_tuple(sample_rate, (KHZ_4, KHZ_16))

    for host_, format_, sample_rate_ in product(
        hosts,
        formats,
        sample_rates,
    ):
        sources.append({
            "host": host_,
            "format": format_,
            "sample_rate": sample_rate_,
            "priority": priority,
        })

    return sources


for klass, fetch in (
    (TimeSeriesBase, fetch_gwosc_data),
    (TimeSeriesBaseDict, fetch_dict),
):
    # register identifier
    klass.get.registry.register_identifier(
        "gwosc",
        klass,
        identify_gwosc_sources,
    )
    # register fetch
    klass.get.registry.register_reader(
        "gwosc",
        klass,
        fetch,
    )



# Copyright (c) 2018-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for `gwpy.plot.colors`."""

import pytest
from matplotlib.colors import (
    LogNorm,
    Normalize,
)
from numpy.testing import assert_array_equal

from .. import colors as plot_colors


@pytest.mark.parametrize(("in_", "factor", "out"), [
    pytest.param(
        "red",
        1.,
        (1., 0., 0.),
        id="red",
    ),
    pytest.param(
        (1., 0., 0.),
        1.,
        (1., 0., 0.),
        id="red-rgb",
    ),
    pytest.param(
        "green",
        .75,
        (0.0, 0.37647058823529411, 0.0),
        id="green",
    ),
])
def test_tint(in_, factor, out):
    """Test `tint()`."""
    assert_array_equal(
        plot_colors.tint(in_, factor=factor),
        out,
    )


def test_format_norm():
    """Test `format_norm()`."""
    norm, kwargs = plot_colors.format_norm({"blah": 1})
    assert isinstance(norm, Normalize)
    assert kwargs == {"blah": 1}


def test_format_norm_log():
    """Test `format_norm()` handling of ``norm="log"``."""
    norm, kwargs = plot_colors.format_norm({
        "norm": "log",
        "vmin": 1,
        "vmax": 10,
    })
    assert isinstance(norm, LogNorm)
    assert norm.vmin == 1
    assert norm.vmax == 10
    assert not kwargs


def test_format_norm_instance():
    """Test `format_norm()` handling of ``norm=<Normalize>``."""
    n = LogNorm()
    norm, kwargs = plot_colors.format_norm({
        "norm": n,
        "clim": (10, 1000),
    })
    assert norm is n
    assert norm.vmin == 10
    assert norm.vmax == 1000
    assert not kwargs


def test_format_norm_clim():
    """Test `format_norm()` handling of ``clim``."""
    norm, kwargs = plot_colors.format_norm({"clim": None})
    assert norm.vmin is None
    assert norm.vmax is None
    assert not kwargs



# Copyright (c) 2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Mock objects for GWpy tests."""

from __future__ import annotations

import inspect
from typing import TYPE_CHECKING
from unittest import mock

from ..detector import Channel
from ..time import LIGOTimeGPS

if TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import TypeAlias

    import nds2
    from astropy.units import UnitBase
    from numpy.typing import NDArray

    from ..timeseries import TimeSeries

    SegmentLike: TypeAlias = tuple[float, float]
    SegmentListLike: TypeAlias = list[tuple[float, float]]

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


# -- NDS2 ----------------------------

def nds2_buffer(
    channel: str,
    data: list | NDArray,
    epoch: LIGOTimeGPS | float,
    sample_rate: float,
    unit: str | UnitBase,
    name: str | None = None,
    slope: float = 1,
    offset: float = 0,
    channel_type: int = 2,
    data_type: int = 8,
) -> nds2.buffer:
    """Create a mocked `nds2.buffer`."""
    import nds2
    epoch = LIGOTimeGPS(epoch)
    ndsbuffer = mock.create_autospec(nds2.buffer)
    ndsbuffer.length = len(data)
    ndsbuffer.channel = nds2_channel(
        channel,
        sample_rate,
        unit,
        channel_type=channel_type,
        data_type=data_type,
    )
    ndsbuffer.name = name or ndsbuffer.channel.name
    ndsbuffer.sample_rate = sample_rate
    ndsbuffer.gps_seconds = epoch.gpsSeconds
    ndsbuffer.gps_nanoseconds = epoch.gpsNanoSeconds
    ndsbuffer.signal_slope = slope
    ndsbuffer.signal_offset = offset
    ndsbuffer.data = data
    return ndsbuffer


def nds2_buffer_from_timeseries(
    ts: TimeSeries,
) -> nds2.buffer:
    """Create a mocked `nds2.buffer` from a :class:`TimeSeries`."""
    return nds2_buffer(
        ts.name or "",
        ts.value,
        ts.x0.value,
        ts.sample_rate.value,
        str(ts.unit),
    )


def nds2_channel(
    name: str,
    sample_rate: float,
    unit: str | UnitBase,
    channel_type: int = 2,
    data_type: int = 8,
) -> nds2.channel:
    """Create a mocked `nds2.channel`."""
    import nds2
    channel = mock.create_autospec(nds2.channel)
    channel.name = name
    channel.sample_rate = sample_rate
    channel.signal_units = str(unit)
    channel.channel_type = channel_type
    channel.channel_type_to_string = nds2.channel.channel_type_to_string
    channel.data_type = data_type
    for attr, value in inspect.getmembers(
        nds2.channel,
        predicate=lambda x: isinstance(x, int),
    ):
        setattr(channel, attr, value)
    return channel


def nds2_connection(
    host: str = "nds.test.gwpy",
    port: int = 31200,
    buffers: Iterable[nds2.buffer] = [],
    protocol: int = 2,
) -> nds2.connection:
    """Create a mock an `nds2.connection` that returns the given buffers."""
    import nds2
    NdsConnection = mock.create_autospec(nds2.connection)  # noqa: N806
    NdsConnection.get_parameter.return_value = False
    NdsConnection.get_host.return_value = host
    NdsConnection.get_port.return_value = int(port)
    NdsConnection.get_protocol.return_value = int(protocol)

    # store buffers internally
    NdsConnection._buffers = list(buffers)  # noqa: SLF001

    def iterate(
        start: float,  # noqa: ARG001
        end: float,  # noqa: ARG001
        names: list[str],
    ) -> list[list[nds2.buffer]]:
        if not NdsConnection._buffers:  # noqa: SLF001
            return []
        return [[
            b for b in NdsConnection._buffers  # noqa: SLF001
            if Channel.from_nds2(b.channel).ndsname in names
        ]]

    NdsConnection.iterate = mock.Mock(side_effect=iterate)

    def find_channels(
        channel_glob: str = "*",
        channel_type_mask: int = nds2.channel.DEFAULT_CHANNEL_MASK,
        data_type_mask: int = nds2.channel.DEFAULT_DATA_MASK,
        min_sample_rate: float = nds2.channel.MIN_SAMPLE_RATE,
        max_sample_rate: float = nds2.channel.MAX_SAMPLE_RATE,
    ) -> list[nds2.channel]:
        out = []
        for b in NdsConnection._buffers:  # noqa: SLF001
            chan = b.channel
            if (
                chan.name == channel_glob
                and chan.sample_rate >= min_sample_rate
                and chan.sample_rate <= max_sample_rate
                and chan.channel_type & channel_type_mask
                and chan.data_type & data_type_mask
            ):
                out.append(chan)
        return out

    NdsConnection.find_channels = mock.Mock(side_effect=find_channels)

    def get_availability(
        names: list[str],
    ) -> list[nds2.availability]:
        out = []
        match = set()
        for buff in NdsConnection._buffers:  # noqa: SLF001
            name = "{0.name},{0.type}".format(Channel.from_nds2(buff.channel))
            if name not in names:
                continue
            start = buff.gps_seconds + buff.gps_nanoseconds * 1e-9
            end = start + buff.length / buff.sample_rate
            segs = [(start, end)]
            out.append(nds2_availability(name, segs))
            match.add(name)
        if missing := match.symmetric_difference(names):
            msg = f"bad channel: {missing.pop()}"
            raise RuntimeError(msg)
        return out

    NdsConnection.get_availability = mock.Mock(side_effect=get_availability)

    return NdsConnection


def nds2_availability(
    name: str,
    segments: SegmentListLike,
) -> nds2.availability:
    """Create a mock `nds2.availability` object."""
    import nds2
    segs = list(map(nds2_segment, segments))
    availability = mock.create_autospec(nds2.availability)
    availability.name = name
    availability.data = segs
    availability.simple_list.return_value = segs
    return availability


def nds2_segment(
    segment: SegmentLike,
) -> nds2.segment:
    """Create a mock `nds2.simple_segment`."""
    import nds2
    nds2seg = mock.create_autospec(nds2.segment)
    nds2seg.gps_start = segment[0]
    nds2seg.gps_stop = segment[1]
    return nds2seg



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for :mod:`gwpy.io.registry`."""

from .. import registry as io_registry

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def test_identify_factory():
    """Test `identify_factory`."""
    id_func = io_registry.identify_factory(".blah", ".blah2")
    assert id_func("read", None, None) is False
    assert id_func("read", "test.txt", None) is False
    assert id_func("read", "test.blah", None) is True
    assert id_func("read", "test.blah2", None) is True
    assert id_func("read", "test.blah2x", None) is False



# Copyright (c) 2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unified I/O read/write for Table objects."""

from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

from astropy.table import vstack
from astropy.table.connect import TableWrite

from ..io.registry import (
    UnifiedFetch,
    UnifiedRead,
    UnifiedWrite,
    default_registry,
)

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Literal

    from .table import EventTable

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


class EventTableRead(UnifiedRead):
    """Read data into an `EventTable`.

    Parameters
    ----------
    source : `str`, `list`
        Source of data, any of the following:

        - `str` path of single data file,
        - `str` path of LAL-format cache file,
        - `list` of paths.

    args
        Other positional arguments will be passed directly to the
        underlying reader method for the given format.

    format : `str`, optional
        The format of the given source files; if not given, an attempt
        will be made to automatically identify the format.

    columns : `list` of `str`, optional
        The list of column names to read.

    where : `str`, or `list` of `str`, optional
        One or more column filters with which to downselect the
        returned table rows as they as read, e.g. ``'snr > 5'``,
        similar to a SQL ``WHERE`` statement.
        Multiple conditions should be connected by ' && ' or ' and ',
        or given as a `list`, e.g. ``'snr > 5 && frequency < 1000'`` or
        ``['snr > 5', 'frequency < 1000']``.

    parallel : `int`
        Number of threads to use for parallel reading of multiple files.

    verbose : `bool`
        Print a progress bar showing read status, default: `False`.

    .. note::

       Keyword arguments other than those listed here may be required
       depending on the `format`

    Returns
    -------
    table : `EventTable`

    Raises
    ------
    astropy.io.registry.IORegistryError
        If the `format` cannot be automatically identified.
    IndexError
        If ``source`` is an empty list.

    Notes
    -----"""

    def merge(  # type: ignore[override]
        self,
        items: Sequence[EventTable],
        join_type: Literal["inner", "exact", "outer"] = "outer",
        metadata_conflicts: Literal["silent", "warn", "error"] = "silent",
    ) -> EventTable:
        """Combine `EventTable` from each file into a single object."""
        return vstack(
            items,
            join_type=join_type,
            metadata_conflicts=metadata_conflicts,
        )

    def __call__(
        self,
        *args,
        **kwargs,
    ) -> EventTable:
        """Read a table."""
        # allow specifying 'where' filter via 'selection' keyword
        if "selection" in kwargs:
            warnings.warn(
                "the 'selection' keyword has been renamed 'where', "
                "this warning will be an error in the near future",
                DeprecationWarning,
                stacklevel=2,
            )
            kwargs.setdefault("where", kwargs.pop("selection"))

        return super().__call__(
            *args,
            **kwargs,
        )


class EventTableWrite(TableWrite, UnifiedWrite):
    """Write this table to a file in the specified format.

    Get help on the available writers for `EventTable` using the
    ``help()`` method:

      >>> EventTable.write.help()  # general help
      >>> EventTable.write.help('root')  # detailed help for the ROOT writer
      >>> EventTable.write.list_formats()  # print list of available formats

    Parameters
    ----------
    target: `str`
        Filename for output data file.

    *args
        Other positional arguments will be passed directly to the
        underlying writer method for the given format.

    format : `str`
        Format for output data; if not given, an attempt will be made
        to automatically identify the format based on the `target`
        filename.

    **kwargs
        Other keyword arguments will be passed directly to the
        underlying writer method for the given format.

    Raises
    ------
    astropy.io.registry.IORegistryError
        If the `format` cannot be automatically identified.

    Notes
    -----"""

    def __init__(self, instance: EventTable, cls: type[EventTable]) -> None:
        """Create a new `EventTableWrite` instance."""
        UnifiedWrite.__init__(
            self,
            instance,
            cls,
            registry=default_registry,
        )


class EventTableFetch(UnifiedFetch):
    """Fetch a table of events from a database.

    Parameters
    ----------
    *args
        All positional arguments are specific to the
        data source, see below for basic usage.

    source : `str`, `~sqlalchemy.engine.Engine`
        The source of the remote data, see _Notes_ for a list of
        registered sources, OR an SQL database `Engine` object.
        Default is ``'sql'``.

    columns : `list` of `str`, optional
        the columns to fetch from the database table, defaults to all

    where : `str`, or `list` of `str`, optional
        One or more column filters with which to downselect the
        returned table rows as they as read, e.g. ``'snr > 5'``;
        multiple conditions should be connected by ' && ', or given as
        a `list`, e.g. ``'snr > 5 && frequency < 1000'`` or
        ``['snr > 5', 'frequency < 1000']``

    engine : `sqlalchemy.engine.Engine`, optional
        The database engine to use when connecting.

    drivername : `str`
        Database backend and driver name.

    user : `str`, optional
        The username for authentication to the database.

    password : `str`, optional
        The password for authentication to the database.

    host : `str`, optional
        The name of the remote database host.

    post : `int`, optional
        Port to connect to on ``host``.

    database : `str`, optional
        The name of the database to connect to.

    query : `dict`, optional
        Query parameters.

    kwargs
        All other positional arguments are specific to the
        data format, see the online documentation for more details.

    Returns
    -------
    table : `EventTable`
        A table of events recovered from the remote database.

    Examples
    --------
    >>> from gwpy.table import EventTable

    To download a table of all blip glitches from the Gravity Spy database:

    >>> EventTable.fetch(
    ...     tablename="glitches",
    ...     source="gravityspy",
    ...     where=["ml_label=Blip", "ml_confidence>0.9"],
    ... )

    To download a table from any SQL-type server

    >>> EventTable.fetch(
    ...     drivername="postgresql",
    ...     host="localname",
    ...     tablename="data",
    ... )

    Notes
    -----"""

    def __call__(
        self,
        *args,
        source: str = "sql",
        **kwargs,
    ):
        # allow specifying 'where' filter via 'selection' keyword
        if "selection" in kwargs:
            warnings.warn(
                "the 'selection' keyword has been renamed 'where', "
                "this warning will be an error in the near future",
                DeprecationWarning,
                stacklevel=2,
            )
            kwargs.setdefault("where", kwargs.pop("selection"))

        # if source is given as a SQL Engine, use it
        if hasattr(source, "connect"):
            kwargs.setdefault("engine", source)
            source = "sql"

        return super().__call__(
            *args,
            source=source,
            **kwargs,
        )



# Copyright (c) 2017-2020 Scott Coughlin
#               2020-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Input/output methods for tabular data.

Access to GravitySpy and O1GlitchClassification triggers requires access
to a PostgresSQL database. Users can set the username and password for
connections in the following environment variables

- ``GRAVITYSPY_DATABASE_USER``
- ``GRAVITYSPY_DATABASE_PASSWORD``

These can be found https://secrets.ligo.org/secrets/144/. The description
is the username and thesecret is the password.
"""

from __future__ import annotations

import os
from typing import TYPE_CHECKING

from .. import EventTable
from . import sql as io_sql

if TYPE_CHECKING:
    from collections.abc import (
        Mapping,
        Sequence,
    )

    import sqlalchemy

    from .. import GravitySpyTable

__author__ = "Scott Coughlin <scott.coughlin@ligo.org>"


def get_gravityspy_triggers(
    # query options
    tablename: str | sqlalchemy.TableClause = "glitches",
    columns: list[str | sqlalchemy.Column] | None = None,
    *,
    where: str | list[str] | None = None,
    order_by: str | None = None,
    order_by_desc: bool = False,
    # connection options
    engine: sqlalchemy.Engine | None = None,
    drivername: str = "postgresql",
    username: str | None = None,
    password: str | None = None,
    host: str = "gravityspyplus.ciera.northwestern.edu",
    port: int = 5432,
    database: str = "gravityspy",
    query: Mapping[str, Sequence[str] | str] | None = None,
    **kwargs,
) -> GravitySpyTable:
    """Fetch data into an `GravitySpyTable`.

    Parameters
    ----------
    tablename : `str`,
        The name of table you are attempting to receive triggers from.

    columns : `list` of `str`, optional
        The columns to read from the database.
        If `None` (default) all columns are read.

    where
        other filters you would like to supply
        underlying reader method for the given format

    order_by : `str`, optional
        Column to order results by.

    order_by_desc : `bool`, optional
        Order results in descending order. Default is `False`.

    engine : `sqlalchemy.engine.Engine`, optional
        An existing SQLAlchemy engine to use for the connection.
        If not given, one will be created using the other connection
        parameters.

    drivername : `str`, optional
        Database backend and driver name.
        Default is ``"postgresql"``.

    username : `str`, optional
        The username for authentication to the database.
        Defaults to the value of the ``GRAVITYSPY_DATABASE_USER``
        environment variable.

    password : `str`, optional
        The password for authentication to the database.
        Defaults to the value of the ``GRAVITYSPY_DATABASE_PASSWD``
        environment variable.

    host : `str`, optional
        The name of the server the database you are connecting to
        lives on.
        Default is ``"gravityspyplus.ciera.northwestern.edu"``

    port : `int`, optional
        Port to connect to on ``host``. Default is ``5432``.

    database : `str`, optional
        The name of the SQL database to connect to.
        Default is ``"gravityspy"``.

    query : `dict`, optional
        Additional query parameters used in the database URL.

    kwargs
        Additional keyword arguments are passed to the
        `EventTable.fetch(..., format="sql")`.

    See Also
    --------
    `EventTable.fetch(source='sql')`
        For details of arguments and keyword arguments for this format.

    Returns
    -------
    table : `GravitySpyTable`
    """
    from sqlalchemy.exc import ProgrammingError

    if engine is None:
        engine = create_engine(
            drivername=drivername,
            username=username,
            password=password,
            host=host,
            port=port,
            database=database,
            query=query,
        )

    try:
        return EventTable.fetch(
            source="sql",
            tablename=tablename,
            columns=columns,
            where=where,
            order_by=order_by,
            order_by_desc=order_by_desc,
            engine=engine,
            **kwargs,
        )
    except ProgrammingError as exc:
        if f'relation "{tablename}" does not exist' in str(exc):
            tables = _get_table_names(engine)
            msg = exc.args[0]
            msg = msg.replace("does not exist", (
                "does not exist, the following tablenames are "
                "acceptable:\n    {}\n".format(
                    "\n    ".join(tables),
                ),
            ))
            exc.args = (msg,)
        raise


# -- utilities -----------------------

def _get_table_names(engine: sqlalchemy.Engine) -> list[str]:
    """Get a list of table names from a SQLAlchemy engine."""
    from sqlalchemy import inspect
    inspector = inspect(engine)
    return inspector.get_table_names()


def create_engine(
    drivername: str = "postgresql",
    username: str | None = None,
    password: str | None = None,
    host: str = "gravityspyplus.ciera.northwestern.edu",
    port: int = 5432,
    database: str = "gravityspy",
    query: Mapping[str, Sequence[str] | str] | None = None,
    **kwargs,
) -> sqlalchemy.Engine:
    """Create a new `sqlalchemy.engine.Engine` for a GravitySpy query.

    This is just a thin wrapper around `sqlalchemy.engine.URL.create`.

    Parameters
    ----------
    drivername : `str, optional
        Database backend and driver name.
        Default is ``"postgresql"``.

    username : `str`, optional
        The username for authentication to the database.
        Defaults to the value of the ``GRAVITYSPY_DATABASE_USER``
        environment variable.

    password : `str`, optional
        The password for authentication to the database.
        Defaults to the value of the ``GRAVITYSPY_DATABASE_PASSWD``
        environment variable.

    host : `str`, optional
        The name of the server the database you are connecting to
        lives on.
        Default is ``"gravityspyplus.ciera.northwestern.edu"``

    port : `int`, optional
        Port to connect to on ``host``. Default is ``5432``.

    database : `str`, optional
        The name of the SQL database your connecting to.
        Default is ``"gravityspy"``.

    query : `dict`, optional
        Additional query parameters used in the database URL.

    kwargs
        Additional keyword arguments are passed to
        :meth:`gwpy.table.io.sql.create_engine`.

    .. note::

       ``username`` and ``passwd`` should be given together, otherwise they
       will be ignored and values will be resolved from the
       ``GRAVITYSPY_DATABASE_USER`` and ``GRAVITYSPY_DATABASE_PASSWD``
       environment variables.

    Returns
    -------
    engine : `sqlalchemy.engine.Engine`
        A new engine.

    See Also
    --------
    gwpy.table.io.sql.create_engine
        For details of how the engine is created.
    """
    if not username:
        username = os.getenv("GRAVITYSPY_DATABASE_USER")
    if not password:
        password = os.getenv("GRAVITYSPY_DATABASE_PASSWD")

    if not username and not password:
        msg = (
            "username and password not given; "
            "remember to either pass or export GRAVITYSPY_DATABASE_USER "
            "and export GRAVITYSPY_DATABASE_PASSWD in order to access the "
            "Gravity Spy Data; LIGO-Virgo-KAGRA members may visit "
            "https://secrets.ligo.org/secrets/144/ for more information."
        )
        raise ValueError(msg)

    return io_sql.create_engine(
        drivername,
        username=username,
        password=password,
        host=host,
        port=port,
        database=database,
        query=query,
        **kwargs,
    )


EventTable.fetch.registry.register_reader(
    "gravityspy",
    EventTable,
    get_gravityspy_triggers,
)
